@@ -231,7 +231,6 @@ class Element(Plugin):
         self.__staged_sources_directory = None  # Location where Element.stage_sources() was called
         self.__tainted = None                   # Whether the artifact is tainted and should not be shared
         self.__required = False                 # Whether the artifact is required in the current session
-        self.__artifact_size = None             # The size of data committed to the artifact cache
 
         # hash tables of loaded artifact metadata, hashed by key
         self.__metadata_keys = {}                     # Strong and weak keys for this key
@@ -1435,15 +1434,20 @@ class Element(Plugin):
             workspace.clear_running_files()
             self._get_context().get_workspaces().save_config()
 
-            # We also need to update the required artifacts, since
-            # workspaced dependencies do not have a fixed cache key
-            # when the build starts.
+            # This element will have already been marked as
+            # required, but we bump the atime again, in case
+            # we did not know the cache key until now.
             #
-            # This does *not* cause a race condition, because
-            # _assemble_done is called before a cleanup job may be
-            # launched.
+            # FIXME: This is not exactly correct, we should be
+            #        doing this at the time which we have discovered
+            #        a new cache key, this just happens to be the
+            #        last place where that can happen.
             #
-            self.__artifacts.append_required_artifacts([self])
+            #        Ultimately, we should be refactoring
+            #        Element._update_state() such that we know
+            #        when a cache key is actually discovered.
+            #
+            self.__artifacts.mark_required_elements([self])
 
     # _assemble():
     #
@@ -1454,6 +1458,9 @@ class Element(Plugin):
     #   - Call the public abstract methods for the build phase
     #   - Cache the resulting artifact
     #
+    # Returns:
+    #    (int): The size of the newly cached artifact
+    #
     def _assemble(self):
 
         # Assert call ordering
@@ -1573,12 +1580,14 @@ class Element(Plugin):
                 }), os.path.join(metadir, 'workspaced-dependencies.yaml'))
 
                 with self.timed_activity("Caching artifact"):
-                    self.__artifact_size = utils._get_dir_size(assembledir)
+                    artifact_size = utils._get_dir_size(assembledir)
                     self.__artifacts.commit(self, assembledir, self.__get_cache_keys_for_commit())
 
             # Finally cleanup the build dir
             cleanup_rootdir()
 
+        return artifact_size
+
     # _pull_pending()
     #
     # Check whether the artifact will be pulled.
@@ -1663,8 +1672,6 @@ class Element(Plugin):
             return False
 
         # Notify successfull download
-        display_key = self._get_brief_display_key()
-        self.info("Downloaded artifact {}".format(display_key))
         return True
 
     # _skip_push():
@@ -1703,16 +1710,13 @@ class Element(Plugin):
             self.warn("Not pushing tainted artifact.")
             return False
 
-        display_key = self._get_brief_display_key()
-        with self.timed_activity("Pushing artifact {}".format(display_key)):
-            # Push all keys used for local commit
-            pushed = self.__artifacts.push(self, self.__get_cache_keys_for_commit())
-            if not pushed:
-                return False
+        # Push all keys used for local commit
+        pushed = self.__artifacts.push(self, self.__get_cache_keys_for_commit())
+        if not pushed:
+            return False
 
-            # Notify successful upload
-            self.info("Pushed artifact {}".format(display_key))
-            return True
+        # Notify successful upload
+        return True
 
     # _shell():
     #
@@ -1817,25 +1821,6 @@ class Element(Plugin):
         workspaces = self._get_context().get_workspaces()
         return workspaces.get_workspace(self._get_full_name())
 
-    # _get_artifact_size()
-    #
-    # Get the size of the artifact produced by this element in the
-    # current pipeline - if this element has not been assembled or
-    # pulled, this will be None.
-    #
-    # Note that this is the size of an artifact *before* committing it
-    # to the cache, the size on disk may differ. It can act as an
-    # approximate guide for when to do a proper size calculation.
-    #
-    # Returns:
-    #    (int|None): The size of the artifact
-    #
-    def _get_artifact_size(self):
-        return self.__artifact_size
-
-    def _get_artifact_cache(self):
-        return self.__artifacts
-
     # _write_script():
     #
     # Writes a script to the given directory.
@@ -2176,7 +2161,8 @@ class Element(Plugin):
     # substituting command strings to be run in the sandbox
     #
     def __extract_variables(self, meta):
-        default_vars = _yaml.node_get(self.__defaults, Mapping, 'variables', default_value={})
+        default_vars = _yaml.node_get(self.__defaults, Mapping, 'variables',
+                                      default_value={})
 
         project = self._get_project()
         if self.__is_junction:
@@ -2189,6 +2175,13 @@ class Element(Plugin):
         _yaml.composite(variables, meta.variables)
         _yaml.node_final_assertions(variables)
 
+        for var in ('project-name', 'element-name', 'max-jobs'):
+            provenance = _yaml.node_get_provenance(variables, var)
+            if provenance and provenance.filename != '':
+                raise LoadError(LoadErrorReason.PROTECTED_VARIABLE_REDEFINED,
+                                "{}: invalid redefinition of protected variable '{}'"
+                                .format(provenance, var))
+
         return variables
 
     # This will resolve the final configuration to be handed
